import inspect
import logging
import os
import re
import time
import warnings
from datetime import datetime
from enum import Enum
from functools import reduce, wraps
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar

import humps
import pandas as pd

from mstrio import __version__ as mstrio_version
from mstrio import config
from mstrio.api.exceptions import (
    MstrTimeoutError,
    PromptedContentError,
    VersionException,
)
from mstrio.types import ObjectSubTypes
from mstrio.utils.dict_filter import filter_list_of_dicts
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.time_helper import (
    DatetimeFormats,
    map_datetime_to_str,
    map_str_to_datetime,
)

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.project_objects.datasets import OlapCube, SuperCube
    from mstrio.server import Project
    from mstrio.types import ObjectTypes

logger = logging.getLogger(__name__)


def deprecation_warning(
    deprecated: str,
    new: str,
    version: str,
    module: bool = True,
    change_compatible_immediately=True,
):
    """This function is used to provide a user with a warning, that a given
    functionality is now deprecated, and won't be supported from a given
    version.

    Args:
        deprecated (str): name of a functionality that is deprecated
        new (str): name of a functionality that replaces deprecated one
        version (str): version from which deprecated functionality won't be
        supported
        module (bool, optional): Whether deprecated functionality is a module.
        Defaults to True.
        change_compatible_immediately (bool, optional): Whether the new
        functionality can be used immediately. Defaults to True.
    """

    module = " module" if module else ""
    if change_compatible_immediately:
        msg = (
            f"{deprecated}{module} is deprecated and will not be supported starting "
            f"from mstrio-py {version}. Please use {new} instead."
        )
    else:
        msg = (
            f"From version {version} {deprecated}{module} will be removed and replaced "
            f"with {new}"
        )
    warnings.warn(DeprecationWarning(msg))


def url_check(url):
    """Checks the validity of the url required in the connection object and
    returns a validated url."""
    regex = r'^https?://.+$'
    match = re.search(regex, url)
    api_index = url.find('/api')

    if match is None:
        msg = (
            "Please check the validity of the base_url parameter. Typically of the "
            "form 'https://<<MSTR Domain>>/MicroStrategyLibrary/'"
        )
        raise ValueError(msg)
    if api_index != -1:
        url = url[:api_index]
    if url.endswith('/'):
        url = url[:-1]

    return url


def version_cut(version):
    res = ".".join([str(int(i)) for i in version.split(".")])
    return res[:6]


def get_args_from_func(func: Callable[[Any], Any]):
    signature = inspect.signature(func)
    return list(signature.parameters.keys())


def get_default_args_from_func(func: Callable[[Any], Any]):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def camel_to_snake(
    response: dict | list,
    whitelist: list[str] = None,
) -> dict | list[dict]:
    """Converts dictionary keys from camelCase to snake_case.
    It works recursively for dicts in dicts."""

    whitelist = whitelist or []

    def convert_dict(source):
        return {
            humps.decamelize(key): value
            if not isinstance(value, dict) or key in whitelist
            else convert_dict(value)
            for key, value in source.items()
        }

    if type(response) == list:
        return [convert_dict(source) for source in response if type(source) == dict]
    elif type(response) == dict:
        return convert_dict(response)
    else:
        raise ValueError("Not supported data type for camel_to_snake conversion")


def snake_to_camel(
    response: dict | list,
    whitelist: list[str] = None,
) -> dict | list[dict]:
    """Converts dictionary keys from snake_case to camelCase.
    It works recursively for dicts in dicts."""
    whitelist = whitelist or []

    def convert_dict(source):
        return {
            humps.camelize(key): value
            if not isinstance(value, dict) or key in whitelist
            else convert_dict(value)
            for key, value in source.items()
        }

    if type(response) == list:
        return [convert_dict(source) for source in response if type(source) == dict]
    elif type(response) == dict:
        return convert_dict(response)
    else:
        raise ValueError("Not supported data type for snake_to_camel conversion")


def check_duplicated_column_names(data_frame):
    list_of_columns = list(data_frame.columns)
    return len(list_of_columns) == len(set(list_of_columns))


def exception_handler(msg, exception_type=Exception, stack_lvl=2):
    """Generic error message handler.

    Args:
        msg (str): Message to print in the Exception or Warning
        exception_type (Exception): Instance of Exception or Warning class
        stack_lvl (int, optional): controls how deep the stacktrace will be
    """
    if not isinstance(exception_type, type) or not issubclass(
        exception_type, Exception
    ):
        raise ValueError("exception_type has to be a subclass of Exception class")
    elif issubclass(exception_type, Exception) and not issubclass(
        exception_type, Warning
    ):
        raise exception_type(msg)
    elif issubclass(exception_type, Warning):
        warnings.warn(msg, exception_type, stacklevel=stack_lvl)


class IServerError(IOError):
    def __init__(self, message, http_code):
        super().__init__(message)
        self.http_code = http_code


def response_handler(response, msg, throw_error=True, verbose=True, whitelist=None):
    """Generic error message handler for transactions against I-Server.

    Args:
        response: Response object returned by HTTP request.
        msg (str): Message to print in addition to any server-generated error
            message(s)
        throw_error (bool): Flag indicates if the error should be thrown
        verbose (bool, optional): controls if messages/errors will be printed
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]
    """
    if whitelist is None:
        whitelist = []

    try:
        logger.debug(f"{response} url = '{response.url}'")
        logger.debug(f"headers = {response.headers}")
        logger.debug(f"content = {response.text}")
        res = response.json()

        if res.get('errors') is not None:
            res = res.get('errors')
            if len(res) > 0:
                res = res[0]

        server_code = res.get('code')
        server_msg = res.get('message')
        ticket_id = res.get('ticketId')
        iserver_code = res.get('iServerCode')
        is_whitelisted = (server_code, response.status_code) in whitelist

        if not is_whitelisted:
            if (
                server_code == 'ERR004'
                and response.status_code == 404
                and server_msg == 'HTTP 404 Not Found'
            ) or (
                server_code == 'ERR001'
                and response.status_code == 405
                and server_msg == 'HTTP 405 Method Not Allowed'
            ):
                msg = (
                    "This REST API functionality is not yet supported on this version "
                    f"of the I-Server: {version_cut(config.iserver_version)}. Please "
                    "upgrade the I-Server version or downgrade the mstrio-py package "
                    f"(current version: {version_cut(mstrio_version)}) in order for "
                    "the versions to match."
                )
                exception_handler(msg, exception_type=VersionException)
            elif iserver_code == -2147206497:  # MSI_REQUEST_TIMEOUT on server-side
                raise MstrTimeoutError(res)
            elif iserver_code == -2147468903:
                raise PromptedContentError(
                    'Prompted content subscription creation is not supported.'
                )
            if verbose:
                logger.error(
                    f'{msg}\n'
                    f'I-Server Error {server_code}, {server_msg}\n'
                    f'Ticket ID: {ticket_id}'
                )
            if throw_error:
                raise IServerError(
                    message=(
                        f"{server_msg}; code: '{server_code}', ticket_id: '{ticket_id}'"
                    ),
                    http_code=response.status_code,
                )
    except JSONDecodeError:
        logger.debug(f"Response body: {response.text}")

        if verbose:
            logger.error(
                f"{msg}\n"
                f"Could not decode the response from the I-Server. Please check if "
                f"I-Server is running correctly"
            )
            # raise error if I-Server response cannot be decoded
            response.raise_for_status()


def fallback_on_timeout(min_limit: int = 50):
    """Return a decorator, which decorates a function with one argument,
    `limit`, to retry with half as big limit if it encounters a timeout error.

    Do note that the wrapped function will return a (result, working_limit)
    tuple, not just the result, to allow for using the working limit further.

    Args:
        min_limit: The minimum limit that will be attempted as retry. If the
            wrapper recurses past it, it will just raise the error back.
    """

    def decorate(func: Callable[[int], Any]):
        """Replaces `fot_wrapper`'s __name__, __doc__ and __module__ with those
        of `func`."""

        @wraps(func)
        def fot_wrapper(limit: int) -> tuple[Any, int]:
            try:
                return func(limit), limit
            except MstrTimeoutError as err:
                new_limit = limit // 2
                if new_limit >= min_limit:
                    logger.warning(
                        f"Timeout hit when executing {func.__name__} with limit "
                        f"{limit}, retrying with limit {new_limit}"
                    )
                    return fot_wrapper(new_limit)
                else:
                    raise err

        return fot_wrapper

    return decorate


def get_parallel_number(total_chunks):
    """Returns the optimal number of threads to be used for downloading
    cubes/reports in parallel."""

    threads = min(8, os.cpu_count() + 4)
    if total_chunks > 0:
        threads = min(total_chunks, threads)

    return threads


def _prepare_objects(
    objects: dict | list[dict],
    filters: Optional[dict] = None,
    dict_unpack_value: Optional[str] = None,
):
    if type(objects) is dict and dict_unpack_value:
        objects = objects[dict_unpack_value]
    objects = camel_to_snake(objects)
    if filters:
        objects = filter_list_of_dicts(objects, **filters)  # type: ignore
    return objects


def fetch_objects_async(
    connection: "Connection",
    api: Callable,
    async_api: Callable,
    limit: Optional[int],
    chunk_size: int,
    filters: dict,
    error_msg: Optional[str] = None,
    dict_unpack_value: Optional[str] = None,
    **kwargs,
) -> list:
    """Get all objects asynchronously. Optionally filter the objects using
    `filters` parameter. Works only for endpoints with `limit` and `offset`
    query parameter (pagination).

    Args:
        connection: MicroStrategy REST API connection object
        api: GET API wrapper function that will return list of objects in bulk
        async_api: asynchronous wrapper of the `api` function
        dict_unpack_value: if the response needs to be unpacked to get into
            the values, specify the keyword
        limit: cut-off value for the number of objects returned
        chunk_size: number of objects in each chunk
        error_msg: specifies error_msg for failed requests
        **filters: dict that specifies filter expressions by which objects will
            be filtered locally
        kwargs: all specific parameters that the api methods require that need
            to be additionally specified
    """
    validate_param_value('limit', limit, int, min_val=1, special_values=[None])
    offset = 0
    chunk_size = min(limit, chunk_size) if limit else chunk_size
    all_objects = []

    # Extract parameters of the api wrapper and set them using the kwargs
    args = get_args_from_func(api)
    param_value_dict = {
        arg: kwargs.get(arg)
        for arg in args
        if arg not in ['connection', 'limit', 'offset', 'error_msg']
    }
    response = api(
        connection=connection,
        offset=offset,
        limit=chunk_size,
        error_msg=error_msg,
        **param_value_dict,
    )
    objects = _prepare_objects(response.json(), filters, dict_unpack_value)
    all_objects.extend(objects)
    current_count = offset + chunk_size
    total_objects = int(response.headers.get('x-mstr-total-count'))
    total_objects = min(limit, total_objects) if limit else total_objects

    if total_objects > current_count:
        it_total = int(total_objects / chunk_size) + (total_objects % chunk_size != 0)
        threads = get_parallel_number(it_total)
        with FuturesSessionWithRenewal(
            connection=connection, max_workers=threads
        ) as session:
            # Extract parameters of the api wrapper and set them using kwargs
            param_value_dict = auto_match_args(
                api,
                kwargs,
                exclude=[
                    'connection',
                    'limit',
                    'offset',
                    'future_session',
                    'error_msg',
                ],
            )
            futures = [
                async_api(
                    future_session=session,
                    connection=connection,
                    offset=offset,
                    limit=chunk_size,
                    **param_value_dict,
                )
                for offset in range(current_count, total_objects, chunk_size)
            ]

        for f in futures:
            response = f.result()
            if not response.ok:
                response_handler(response, error_msg, throw_error=False)
            objects = _prepare_objects(response.json(), filters, dict_unpack_value)
            all_objects.extend(objects)
    return all_objects


def fetch_objects(
    connection: "Connection",
    api: Callable,
    limit: Optional[int],
    filters: dict,
    error_msg: Optional[str] = None,
    dict_unpack_value: Optional[str] = None,
    **kwargs,
) -> list:
    """Fetch and prepare objects. Optionally filter the objects by using the
    filters parameter. This function only supports endpoints without pagination.

    Args:
        connection: MicroStrategy REST API connection object
        api: GET API wrapper function that will return list of objects in bulk
        dict_unpack_value: if the response needs to be unpacked to get into
            the values specify the keyword
        limit: cut-off value for the number of objects returned
        error_msg: specifies error_msg for failed requests
        **filters: dict that specifies filter expressions by which objects will
            be filtered locally
        kwargs: all specific parameters that the api methods require that need
            to be additionally specified
    """
    validate_param_value('limit', limit, int, min_val=1, special_values=[None])

    # Extract parameters of the api wrapper and set them using kwargs
    param_value_dict = auto_match_args(api, kwargs, exclude=['connection', 'error_msg'])
    response = api(connection=connection, error_msg=error_msg, **param_value_dict)
    if response.ok:
        objects = _prepare_objects(response.json(), filters, dict_unpack_value)
        if limit:
            objects = objects[:limit]
        return objects
    else:
        return []


def sort_object_properties(source: dict) -> int:
    """Sort all properties of an object representing an MSTR object."""
    preffered_order = {
        'id': 1,
        'name': 2,
        'description': 3,
        'alias': 4,
        'type': 5,
        'subtype': 6,
        'ext_type': 7,
        'acg': 101,
        'acl': 102,
    }
    return preffered_order.get(source, 50)


def auto_match_args(
    func: Callable,
    param_dict: dict,
    exclude: Optional[list] = None,
    include_defaults: bool = True,
) -> dict:
    """Automatically match dict data to function arguments.

    Handles default parameters. Extracts value from Enums. Returns matched
    arguments as dict.

    Note: don't use it for alter purposes as, changing parameter value
        back to default currently doesn't work (Rework line 413?)

    Args:
        function: function for which args will be matched
        param_dict: dict to use for matching the function args
        exclude: set `exclude` parameter to exclude specific param-value pairs
        include_defaults: if `False` then values which have the same value as
            default will not be included in the result
    Raises:
        KeyError: could not match all required arguments
    """

    exclude = exclude or []
    args = get_args_from_func(func)
    default_dict = get_default_args_from_func(func)

    param_value_dict = {}
    for arg in args:
        if arg in exclude:
            continue
        else:
            val = (
                param_dict.get(arg)
                if param_dict.get(arg) is not None
                else default_dict.get(arg)
            )
            if (
                not include_defaults
                and arg in default_dict
                and val == default_dict[arg]
            ):
                continue

            if isinstance(val, Enum):
                val = val.value
        param_value_dict.update({arg: val})

    return param_value_dict


def flatten2list(object) -> list:
    """Flatten to list nested objects of type list, tuple, sets."""
    gather = []
    for item in object:
        if isinstance(item, (list, tuple, set)):
            gather.extend(flatten2list(item))
        else:
            gather.append(item)
    return gather


def dict_compare(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys
    modified = {o: (d1[o], d2[o]) for o in intersect_keys if d1[o] != d2[o]}
    same = {o for o in intersect_keys if d1[o] == d2[o]}
    return added, removed, modified, same


def __validate_single_param_value(
    value,
    param_name,
    data_type,
    max_val,
    min_val,
    regex,
    valid_example,
    inv_val,
    special_values=None,
):
    special_values = special_values or []
    if value in special_values:
        return True

    if max_val is not None and value > max_val:
        msg = f"'{param_name}' has to be less than or equal to {max_val}"
        exception_handler(msg, inv_val)
        return False
    elif min_val is not None and value < min_val:
        msg = f"'{param_name}' has to be greater than or equal to {min_val}"
        exception_handler(msg, inv_val)
        return False
    elif regex is not None and not re.match(regex, value):
        pattern = valid_example or regex
        msg = f"'{param_name}' has to match pattern '{pattern}'"
        exception_handler(msg, inv_val)
        return False
    elif (
        all(cond is None for cond in [max_val, min_val, regex])
        and special_values
        and str not in data_type
    ):
        msg = f"'{param_name}' has to be one of {special_values}"
        exception_handler(msg, inv_val)
        return False
    return True


def validate_param_value(
    param_name,
    param_val,
    data_type,
    max_val=None,
    min_val=None,
    special_values=None,
    regex=None,
    exception=True,
    valid_example=None,
) -> bool:
    """Validate param data type and optionally max, min special values.

    Raise:
        TypeError
        ValueError
    """
    special_values = special_values or []
    inv_type = TypeError if exception else Warning
    inv_val = ValueError if exception else Warning
    data_type = data_type if isinstance(data_type, list) else [data_type]

    if any(
        map(lambda x: x == param_val and isinstance(x, type(param_val)), special_values)
    ):
        return True

    if type(param_val) not in data_type:
        msg = f"'{param_name}' needs to be of type {data_type}"
        exception_handler(msg, inv_type)
        return False
    if type(param_val) == list:
        return all(
            [
                __validate_single_param_value(
                    value,
                    param_name,
                    data_type,
                    max_val,
                    min_val,
                    regex,
                    valid_example,
                    inv_val,
                    special_values,
                )
                for value in param_val
            ]
        )

    return __validate_single_param_value(
        param_val,
        param_name,
        data_type,
        max_val,
        min_val,
        regex,
        valid_example,
        inv_val,
        special_values,
    )


def extract_all_dict_values(list_of_dicts: list[dict]) -> list[Any]:
    """Extract list of dicts values into list."""
    all_options = []
    for option in list_of_dicts:
        all_options.extend(option.values())
    return all_options


def delete_none_values(
    source: dict, *, whitelist_attributes: Optional[list] = None, recursion: bool
) -> dict:
    """Delete keys with None values from dictionary.

    Args:
        source (dict): dict object from which none values will be deleted
        whitelist_attributes (list): keyword-only argument containing name
            of attribute, which should be left alone even if they contain
            none values
        recursion (bool): flag that turns recursion on or off
    """
    whitelist = whitelist_attributes or []

    new_dict = {}
    for key, value in source.items():
        if recursion and isinstance(value, dict):
            new_dict[key] = delete_none_values(
                value, whitelist_attributes=whitelist, recursion=recursion
            )
        elif value not in [[], {}, None] or key in whitelist:
            new_dict[key] = value

    return new_dict


def get_objects_id(obj, obj_class):
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, obj_class):
        return obj.id
    return None


def list_folders(
    connection,
    name: Optional[str] = None,
    to_dataframe: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> list[dict] | pd.DataFrame:
    """List folders.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        name (string, optional): exact name of a folder to filter results.
            If `None`, all folders will be retrieved.
        to_dataframe (bool, optional): determine if result of retrieval will
            be returned as a list of dicts or DataFrame.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'subtype',
            'type','date_created', 'date_modified', 'version', 'acg', 'owner']

    Returns:
        When `to_dataframe` set to False (default value) then DataFrame with
        folders, otherwise list of dictionaries with folders.
    """
    from mstrio.api import objects

    DSS_XML_SEARCH_TYPE_EXACTLY = 2
    FOLDER_TYPE = 8

    msg = "Error while creating an instance for searching objects"
    res_e = objects.create_search_objects_instance(
        connection=connection,
        name=name,
        pattern=DSS_XML_SEARCH_TYPE_EXACTLY,
        object_type=FOLDER_TYPE,
        error_msg=msg,
    )
    search_id = res_e.json()['id']
    msg = "Error while retrieving folders from the environment."
    fldrs = fetch_objects_async(
        connection,
        api=objects.get_objects,
        async_api=objects.get_objects_async,
        limit=limit,
        chunk_size=1000,
        error_msg=msg,
        filters=filters,
        search_id=search_id,
    )

    if to_dataframe:
        return pd.DataFrame(fldrs)
    else:
        return fldrs


def create_folder(
    connection,
    folder_name: str,
    folder_description: Optional[str] = None,
    parent_name: Optional[str] = None,
    parent_id: Optional[str] = None,
):
    """Create a folder.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        folder_name (string): name of the folder which will be created
        folder_description (string, optional): description of the folder which
            will be created
        parent_name (string, optional): name of the folder in which new folder
            will be created
        parent_id (string, optional): id of the folder in which new folder
            will be created
        error_msg (string, optional): error message

    Returns:
        full response about the creation of the folder

    Raises:
        Exception when:
            - neither `parent_name` nor `parent_id` is specified
            - both `parent_name` and `parent_id` are specified
            - parent folder was not found or multiple folders with the same
              name exists when providing `parent_name`
    """
    from mstrio.api import folders

    if parent_id is None and parent_name is None:
        exception_handler("Please specify either 'parent_name' or 'parent_id'.")
    if parent_name is not None and parent_id is not None:
        exception_handler(
            "Please specify either 'parent_name' or 'parent_id' but not both."
        )
    if parent_id is None:
        fldrs = list_folders(connection=connection, name=parent_name)
        if len(fldrs) != 1:
            msg = f"Parent folder with name '{parent_name}' was either not found "
            msg += "or multiple folders with the same name exists."
            exception_handler(msg)
        parent_id = fldrs[0]['id']
    return folders.create_folder(
        connection=connection,
        name=folder_name,
        parent_id=parent_id,
        description=folder_description,
    )


def delete_folder(
    connection, id: Optional[str] = None, name: Optional[str] = None, error_msg=None
):
    """Delete a folder.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string, optional): id of the folder which will be deleted
        name (string, optional): name of the folder which will be deleted
        error_msg (string, optional): error message

    Returns:
        full response about the deletion of the folder

    Raises:
        Exception when:
            - neither `name` nor `id` is specified
            - both `name` and `id` are specified
            - folder was not found or multiple folders with the same name
              exists when providing `name`
    """
    from mstrio.api import objects

    if id is None and name is None:
        exception_handler(
            "Please specify either 'name' or 'id' of the folder to be deleted."
        )
    if name is not None and id is not None:
        msg = (
            "Please specify either 'name' or 'id' of the folder to be deleted "
            "but not both."
        )
        exception_handler(msg)
    if id is None:
        fldrs = list_folders(connection=connection, name=name)
        if len(fldrs) != 1:
            msg = f"Folder with name '{name}' was either not found "
            msg += "or multiple folders with the same name exists."
            exception_handler(msg)
        id = fldrs[0]['id']
    FOLDER_TYPE = 8
    return objects.delete_object(
        connection=connection, id=id, object_type=FOLDER_TYPE, error_msg=error_msg
    )


def merge_id_and_type(
    object_id: str,
    object_type: 'ObjectTypes | ObjectSubTypes | int',
    error_msg: Optional[str] = None,
) -> str:
    if not object_id or not object_type:
        exception_handler(
            msg=error_msg or "Please provide both `id` and `type`.",
            exception_type=AttributeError,
        )
    object_id = get_objects_id(object_id, type(object_id))
    object_type = (
        get_enum_val(object_type, type(object_type))
        if isinstance(object_type, Enum)
        else object_type
    )
    return f'{object_id};{object_type}'


def rsetattr(obj, attr, val):
    """Recursive setattr. This can only modify last attr in the chain.
    For example rsetattr(obj, 'attr1.attr2.attr3', new_value) won't work
    if attr2 doesn't exist.

    Args:
        obj: An object that is edited
        attr: attribute name chain/path
        val: new value
    Returns:
        None on success
    Raises:
        AttributeError on failure, when the 'attr' is incorrect
    Example:
        rsetattr(obj, 'attr1.attr2', new_value)
    """

    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *default):
    """Recursive getattr. Third parameter is the default value.

    Args:
        obj: An object that is inspected
        attr: attribute name chain/path
        default (optional): value returned if 'attr' not found
    Returns:
        value of 'attr' on success. On failure, default value.
    Raises:
        AttributeError on failure, when the 'attr' is incorrect
        and no default provided
    Example:
        rgetattr(obj, 'attr1.attr2', 'default value')
    """

    def _getattr(obj, attr):
        return getattr(obj, attr, *default)

    return reduce(_getattr, [obj] + attr.split('.'))


def filter_params_for_func(
    func: Callable, params: dict, exclude: Optional[list] = None
) -> dict:
    """Filter dict of parameters and return only those that are parameters
    of a `func`.
    Mainly used in `EntityBase.alter()`, before calling
    `EntityBase._alter_properties()` as shown in `Example`.

    Args:
        func: a function that the params are going to be fit for
        params: a dict with parameters that will be filtered
        exclude: set `exclude` parameter to exclude specific param-value pairs
    Returns:
        a dict with parameters of `func` that exist in `params` dict
    Example:
        params = filter_params_for_func(self.alter, locals())
        self._alter_properties(**params)
    """
    args = get_args_from_func(func)
    defaults_dict = get_default_args_from_func(func)
    exclude = exclude or []
    properties = {}

    for arg in args:
        if arg in exclude:
            continue
        elif params.get(arg, None) is not None:
            properties.update({arg: params.get(arg)})
        elif defaults_dict.get(arg, None) is not None:
            properties.update({arg: defaults_dict.get(arg)})
    return properties


T = TypeVar("T")


def filter_obj_list(obj_list: list[T], **filters: dict[str, Any]) -> list[T]:
    """
    Filter a list of objects by providing one or more key-value pair filters.
    """
    return [
        obj for obj in obj_list if all(getattr(obj, f) == v for f, v in filters.items())
    ]


def choose_cube(
    connection: 'Connection', cube_dict: dict
) -> 'OlapCube | SuperCube | None':
    """Return correct cube object based on dictionary with cube's info.

    Note: In case of wrong subtype, `None` is returned.
    """
    cube_subtype = cube_dict['subtype']
    if cube_subtype == ObjectSubTypes.OLAP_CUBE.value:
        from mstrio.project_objects.datasets.olap_cube import OlapCube

        return OlapCube.from_dict(cube_dict, connection)
    elif cube_subtype == ObjectSubTypes.SUPER_CUBE.value:
        from mstrio.project_objects.datasets.super_cube import SuperCube

        return SuperCube.from_dict(cube_dict, connection)


def get_valid_project_id(
    connection: "Connection",
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    with_fallback: bool = False,
):
    """Check if the project name exists and return the project ID.

    Args:
        connection(object): MicroStrategy connection object
        project_id: Project ID
        project_name: Project name
        with_fallback: Specify if the project should be taken from `connection`
        object if `project_id` is not specified and the project failed to be
        found based on `project_name`
    """
    from mstrio.server import Project

    # Search for a project by its name if id was not specified, but name was
    if not project_id:
        project_loaded_list = (
            Project._list_loaded_projects(
                connection, to_dictionary=True, name=project_name
            )
            if project_name
            else []
        )
        if project_loaded_list:
            project_id = project_loaded_list[0]['id']
        else:
            if project_name:
                msg = (
                    f"There is no project with the given name: '{project_name}'"
                    f" or the project is not loaded."
                )
            else:
                msg = "`project_id` and `project_name` were not provided. "
            if with_fallback:
                logger.info(msg + "Project from `connection` object is used instead.")
                project_id = fallback_to_conn_project_id(connection)
            else:
                exception_handler(
                    msg + "Please specify valid `project_id` or `project_name`",
                    exception_type=ValueError,
                )

    return project_id


def fallback_to_conn_project_id(connection: "Connection") -> Optional[str]:
    try:
        connection._validate_project_selected()
        return connection.project_id
    except AttributeError:
        exception_handler(
            msg="Project could not be determined.", exception_type=ValueError
        )


def get_valid_project_name(connection: "Connection", project_id: str):
    """Returns project name of given project based on its ID.

    Args:
        connection(object): MicroStrategy connection object
        project_id: Project ID
    """
    from mstrio.server import Project

    project_loaded_list = Project._list_loaded_projects(
        connection, to_dictionary=True, id=project_id
    )
    project_name = project_loaded_list[0]['name']

    return project_name


class Dictable:
    """The fundamental class in mstrio-py package. Includes support for
    converting an object to a dictionary, and creating an object from a
    dictionary."""

    _FROM_DICT_MAP: dict[str, Callable] = {}  # map attributes to Enums and components
    # list of attribute name, which are allowed to have none values
    # in dict returned by .to_dict()
    _ALLOW_NONE_ATTRIBUTES: list[str] = []
    _KEEP_CAMEL_CASE: list[str] = []

    @classmethod
    def _unpack_objects(cls, key, val, camel_case=True):
        """Unpack Enums, Dictable obj, list of Enums"""
        if isinstance(val, datetime):
            return map_datetime_to_str(key, val, cls._FROM_DICT_MAP)
        elif isinstance(val, Enum):
            return val.value
        elif isinstance(val, Dictable):
            return val.to_dict(camel_case)
        elif isinstance(val, list):
            return [cls._unpack_objects(key, v, camel_case) for v in val]
        return val

    @classmethod
    def _dict_to_obj(cls, connection, val, key):
        def constructor():
            if isinstance(cls._FROM_DICT_MAP[key], DatetimeFormats):
                return map_str_to_datetime(key, val, cls._FROM_DICT_MAP)
            elif isinstance(cls._FROM_DICT_MAP[key], type(Enum)):
                return cls._FROM_DICT_MAP[key](val)
            elif isinstance(cls._FROM_DICT_MAP[key], type(Dictable)):
                return cls._FROM_DICT_MAP[key].from_dict(val)
            elif isinstance(cls._FROM_DICT_MAP[key], list):
                if isinstance(cls._FROM_DICT_MAP[key][0], list):
                    # for: List[List[handling_cls]]
                    handling_cls = cls._FROM_DICT_MAP[key][0][0]
                    return [[handling_cls.from_dict(item) for item in v] for v in val]
                elif (
                    all(isinstance(v, type(Enum)) for v in cls._FROM_DICT_MAP[key])
                    and val is not None
                ):
                    return [cls._FROM_DICT_MAP[key][0](v) for v in val]
                elif all(
                    [isinstance(v, type(Dictable)) for v in cls._FROM_DICT_MAP[key]]
                ):
                    return [cls._FROM_DICT_MAP[key][0].from_dict(v) for v in val]
                elif callable(cls._FROM_DICT_MAP[key][0]):
                    return [cls._FROM_DICT_MAP[key][0](v, connection) for v in val]
            else:
                return cls._FROM_DICT_MAP[key](val, connection)

        return val if key not in cls._FROM_DICT_MAP else constructor()

    def to_dict(self, camel_case: bool = True) -> dict:
        """Converts an object to a dictionary excluding object's private
        properties. When converting the object to a dictionary, the object's
        attributes become the dictionary's keys and are in camel case by default
        Attribute values stored as objects are automatically converted to
        non-/ primitive data structures.

        Args:
            camel_case (bool, optional): Set to True if attribute names should
                be converted from snake case to camel case. Defaults to True.

        Returns:
            dict: A dictionary representation of object's attributes and values.
                By default, the dictionary keys are in camel case.
        """

        hidden_keys = [
            '_fetched_attributes',
            '_altered_properties',
            '_connection',
            'connection',
            '_type',
        ]
        cleaned_dict = self.__dict__.copy()
        properties = {
            elem[0]
            for elem in inspect.getmembers(
                self.__class__, lambda x: isinstance(x, property)
            )
        }
        for prop in properties:
            to_be_deleted = '_' + prop
            cleaned_dict[prop] = cleaned_dict.pop(to_be_deleted, None)
        result = {
            key: self._unpack_objects(key, val, camel_case)
            for key, val in cleaned_dict.items()
            if key not in hidden_keys
        }

        result = delete_none_values(
            result, whitelist_attributes=self._ALLOW_NONE_ATTRIBUTES, recursion=False
        )
        result = {
            key: result[key] for key in sorted(result, key=sort_object_properties)
        }
        return (
            snake_to_camel(result, whitelist=self._KEEP_CAMEL_CASE)
            if camel_case
            else result
        )

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Optional["Connection"] = None,
        to_snake_case: bool = True,
    ) -> T:
        """Creates an object from a dictionary. The dictionary's keys in camel
        case are changed to object's attribute names (by default in snake case)
        and dict values are composed to their proper data types such as Enums,
        list of Enums etc. as specified in _FROM_DICT_MAP.

        Args:
            cls (T): Class (type) of an object that should be created.
            source (Dict[str, Any]): A dictionary from which an object will be
                constructed.
            connection (Connection, optional): A MSTR Connection object.
                Defaults to None.
            to_snake_case (bool, optional): Set to True if attribute names
                should be converted from camel case to snake case. Defaults to
                True.

        Returns:
            T: An object of type T.
        """
        object_source = (
            camel_to_snake(source, whitelist=cls._KEEP_CAMEL_CASE)
            if to_snake_case
            else source
        )

        if connection is not None:
            object_source["connection"] = connection

        args = {
            key: cls._dict_to_obj(connection, val, key)
            for key, val in object_source.items()
            if key in cls.__init__.__code__.co_varnames
        }
        obj = cls(**args)  # type: ignore
        return obj

    @classmethod
    def bulk_from_dict(
        cls: T,
        source_list: list[dict[str, Any]],
        connection: Optional["Connection"] = None,
        to_snake_case: bool = True,
    ) -> list[T]:
        """Creates multiple objects from a list of dictionaries. For each
        dictionary provided the keys in camel case are changed to object's
        attribute names (by default in snake case) and dict values are composed
        to their proper data types such as Enums, list of Enums etc. as
        specified in the object's _FROM_DICT_MAP.

        Args:
            cls (T): Class (type) of the objects that should be created.
            source_list (List[Dict[str, Any]]): A list of dictionaries from
                which the objects will be constructed.
            connection (Connection, optional): A MSTR Connection object.
                Defaults to None.
            to_snake_case (bool, optional): Set to True if attribute names
                should be converted from camel case to snake case. Defaults to
                True.

        Returns:
            T: A list of objects of type T.
        """
        return [
            cls.from_dict(
                source=source, connection=connection, to_snake_case=to_snake_case
            )
            for source in source_list
        ]

    def __repr__(self) -> str:
        from mstrio.utils.entity import auto_match_args_entity

        param_dict = auto_match_args_entity(
            self.__init__,
            self,
            exclude=['self'],
            include_defaults=False,
        )
        formatted_params = ', '.join(
            (f'{param}={repr(value)}' for param, value in param_dict.items())
        )
        return f'{self.__class__.__name__}({formatted_params})'


def is_dossier(view_media: int):
    """Documents and dossiers have the same type and subtype when returned
    from search api. They can be distinguished only by view_media value.
    """
    return (
        view_media & 4160749568 == 1879048192 or view_media & 4160749568 == 1610612736
    )


def is_document(view_media: int):
    """Documents and dossiers have the same type and subtype when returned
    from search api. They can be distinguished only by view_media value.
    """
    return not is_dossier(view_media)


def rename_dict_keys(source: dict, mapping: dict) -> dict:
    """Rename dict keys according to mapping.

    Args:
        source (dict): An original dictionary, which keys are to be renamed.
        mapping (dict): A dictionary containing mapping of keys.

    Returns:
        dict: A dictionary with keys renamed.
    """
    for rest_name, python_name in mapping.items():
        if rest_name in source:
            old = source.pop(rest_name)
            if python_name:
                source[python_name] = old
    return source


def verify_project_status(
    project: 'Project', correct_statuses: list[str] | str, node: Optional[str] = None
) -> bool:
    """Veriy if provided status is correct for given project.

    Args:
        project (Project): Project for which statuses will be verified.
        correct_statuses (list[str], str): A list of correct statuses
            or just one status.
        node (str, optional): Node name on which status should be verified,
            represented as name of first node if not provided.

    Returns:
        bool: True if status is correct and False otherwise.
    """

    def get_status(project: 'Project', node: Optional[str] = None) -> str:
        node_name = project.nodes[0]['name'] if not node else node
        nodes_filtered = [node for node in project.nodes if node['name'] == node_name]

        if not nodes_filtered:
            raise ValueError(
                f"Node {node} not found. Available nodes "
                f"{[node['name'] for node in project.nodes]}"
            )

        projects_filtered = [
            proj
            for node in nodes_filtered
            for proj in node['projects']
            if proj['id'] == project.id
        ]

        return projects_filtered[0]['status']

    status = get_status(project=project, node=node)
    iteration = 0

    while status not in correct_statuses and iteration < 20:
        time.sleep(1)
        project.fetch('nodes')
        status = get_status(project=project, node=node)
        iteration += 1

    return status in correct_statuses



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(
    err_msg='Error obtaining the list of registered nodes from the MicroStrategy '
    'deployment.'
)
def get_nodes(connection, error_msg=None):
    """Obtain the list of registered nodes from the MicroStrategy deployment.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection().
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(url=f'{connection.base_url}/api/registrations/nodes')


@ErrorHandler(
    err_msg='Error obtaining the list of registered services available from '
    'the MicroStrategy deployment'
)
def get_services(connection, error_msg=None):
    """Obtain the list of registered services available from the MicroStrategy
    deployment.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection().
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(url=f'{connection.base_url}/api/registrations/services')


@ErrorHandler(
    err_msg='Error obtaining the metadata information for the registered services'
    ' available from the MicroStrategy deployment.'
)
def get_services_metadata(connection, error_msg=None):
    """Obtain the metadata information for the registered services available
    from the MicroStrategy deployment.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection().
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(
        url=f'{connection.base_url}/api/registrations/services/metadata'
    )


@ErrorHandler(err_msg='Error to start/stop service')
def start_stop_service(
    connection, login, password, name, id, address, action="START", error_msg=None
):
    """Start or stop registered service.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection()
        login (string): login for SSH operation
        password (string): password for SSH operation
        name(string): name of the service
        id(string): name of the service
        action(string): one of "START" or "STOP"
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object.
    """

    body = {
        "name": name,
        "id": id,
        "action": action,
        "address": address,
        "login": login,
        "password": password,
    }
    url = f'{connection.base_url}/api/registrations/services/control'
    response = connection.post(url=url, json=body)
    return response



from enum import auto
from typing import TYPE_CHECKING, Optional

from mstrio.api import objects
from mstrio.connection import Connection
from mstrio.modeling.expression import Expression, FactExpression
from mstrio.modeling.schema.helpers import (
    DataType,
    FormReference,
    SchemaObjectReference,
)
from mstrio.types import ObjectTypes
from mstrio.utils.entity import Entity
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import (
    delete_none_values,
    exception_handler,
    filter_params_for_func,
)

if TYPE_CHECKING:
    from mstrio.modeling.schema.attribute.attribute_form import AttributeForm


class AttributeForm(Entity):  # noqa
    """The Attribute Form Object

    Args:
        connection: a `Connection` object tied to the desired environment
        id: Attribute Form ID. Note that Form ID is associated with the form
            category used. Multiple forms that use the same category will have
            the same ID. However, since no two forms within the same attribute
            can use the same category, no two forms will share an ID in
            the same attribute.
        name: The name of the attribute form set by the attribute. Unlike
            category, which is the systemic name associated with each reusable
            form, this name is specific to the attribute using this form.
        description: description of the AttributeForm
        category: The category of the attribute form. Unlike name, this field
            is independent of the attribute using this form. This field can
            only be set when creating a new form. Once a form is created, its
            category becomes non-mutable. If not provided (or set as None) when
            an attribute is being created, a custom category will be
            automatically generated.
        form_type: A read-only field indicating the type of this form. A custom
            form is created if its category is set to None.
        display_format: display format of the AttributeForm
        data_type: Representation in the object model for a data-type that
            could be used for a SQL column.
        expressions: Array with a member object for each separately defined
            expression currently in use by a fact. Often a fact expression
            takes the form of just a single column name, but more complex
            expressions are possible.
        alias: alias of the AttributeForm
        lookup_table: lookup table of the AttributeForm. It has to be a lookup
            table used in one of the expressions assigned to AttributeForm
        child_forms: only used if 'is_form_group' is set to true
        geographical_role: identifies the type of geographical information this
            form represents
        time_role: time role of the AttributeForm
        is_form_group: A boolean field indicating whether this form is a form
            group (if true) or a simple form (if false).
        is_multilingual: A boolean field indicating whether this field is
            multilingual. Any key form of the attribute is not allowed to be
            set as multilingual.
    """

    class FormType(AutoName):
        """Enumeration constants used to specify a type of this form."""

        CUSTOM = auto()
        SYSTEM = auto()

    class DisplayFormat(AutoName):
        """Enumeration constants used to specify display format of
        the attribute form."""

        NUMBER = auto()
        TEXT = auto()
        PICTURE = auto()
        URL = auto()
        EMAIL = auto()
        HTML_TAG = auto()
        DATE = auto()
        TIME = auto()
        SYMBOL = auto()
        PHONE_NUMBER = auto()
        DATE_TIME = auto()
        BIG_DECIMAL = auto()

    class GeographicalRole(AutoName):
        """Enumeration constants used to specify geographical role of
        the attribute form."""

        NONE = auto()
        CITY = auto()
        STATE = auto()
        COUNTRY = auto()
        LOCATION = auto()
        LATITUDE = auto()
        LONGITUDE = auto()
        OTHER = auto()
        ZIP_CODE = auto()
        COUNTY = auto()
        AREA_CODE = auto()
        GEOMETRY = auto()

    class TimeRole(AutoName):
        """Enumeration constants used to specify time role of
        the attribute form."""

        NONE = auto()
        DATE = auto()
        TIME = auto()
        SECOND = auto()
        MINUTE = auto()
        HOUR = auto()
        DAY = auto()
        WEEK = auto()
        MONTH = auto()
        QUARTER = auto()
        YEAR = auto()
        SEASONAL_ROOT = auto()
        YEAR_OF_DECADE = auto()
        QUARTER_OF_YEAR = auto()
        MONTH_OF_YEAR = auto()
        MONTH_OF_QUARTER = auto()
        WEEK_OF_YEAR = auto()
        WEEK_OF_QUARTER = auto()
        WEEK_OF_MONTH = auto()
        DAY_OF_YEAR = auto()
        DAY_OF_QUARTER = auto()
        DAY_OF_MONTH = auto()
        DAY_OF_WEEK = auto()
        HOUR_OF_DAY = auto()
        MINUTE_OF_DAY = auto()
        MINUTE_OF_HOUR = auto()
        SECOND_OF_DAY = auto()
        SECOND_OF_HOUR = auto()
        SECOND_OF_MINUTE = auto()
        SEASONAL_END = auto()

    _OBJECT_TYPE = ObjectTypes.ATTRIBUTE_FORM
    _API_GETTERS = {
        (
            'id',
            'name',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
        ): objects.get_object_info
    }

    def __init__(self, connection: Connection, id: str) -> None:
        super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.category = kwargs.get('category')
        if self.FormType.has_value(kwargs.get('type')):
            self.form_type = self.FormType(kwargs.get('type'))
        elif kwargs.get('type'):
            self.type = kwargs.get('type')

        self.is_multilingual = kwargs.get('is_multilingual', False)
        self.is_form_group = kwargs.get('is_form_group', False)
        self.geographical_role = (
            self.GeographicalRole(kwargs.get('geographical_role'))
            if kwargs.get('geographical_role')
            else None
        )
        self.time_role = (
            self.TimeRole(kwargs.get('time_role')) if kwargs.get('time_role') else None
        )
        self.display_format = (
            self.DisplayFormat(kwargs.get('display_format'))
            if kwargs.get('display_format')
            else None
        )
        self.data_type = (
            DataType.from_dict(kwargs.get('data_type'))
            if kwargs.get('data_type')
            else None
        )

        self.expressions = (
            [
                FactExpression.from_dict(expr, self._connection)
                for expr in kwargs.get('expressions')
            ]
            if kwargs.get('expressions')
            else None
        )

        self.alias = kwargs.get('alias')
        self.lookup_table = (
            SchemaObjectReference.from_dict(kwargs.get('lookup_table'))
            if kwargs.get('lookup_table')
            else None
        )

        self.child_forms = (
            [
                FormReference.from_dict(expr, self._connection)
                for expr in kwargs.get('child_forms')
            ]
            if kwargs.get('child_forms')
            else None
        )

    @classmethod
    def local_create(
        cls,
        connection: Connection,
        name: str,
        description: Optional[str] = None,
        category: Optional[str] = None,
        display_format: Optional[DisplayFormat] = None,
        data_type: Optional[DataType] = None,
        expressions: Optional[list[FactExpression]] = None,
        alias: Optional[str] = None,
        lookup_table: Optional[SchemaObjectReference] = None,
        child_forms: Optional[list[FormReference]] = None,
        geographical_role: Optional[GeographicalRole] = None,
        time_role: Optional[TimeRole] = None,
        is_form_group: bool = False,
        is_multilingual: bool = False,
    ) -> "AttributeForm":
        """Internal method that creates ONLY LOCAL AttributeForm object.
        In order to create an AttributeForm object on the server,
        use `Attribute.add_form()` method of a corresponding Attribute object.
        """
        properties = filter_params_for_func(
            cls.local_create, locals(), exclude=['connection']
        )
        properties = delete_none_values(properties, recursion=True)

        obj = cls.__new__(cls)
        for key, val in properties.items():
            setattr(obj, key, val)
        obj._id = None
        obj._connection = connection
        return obj

    def list_properties(self, camel_case=True) -> dict:
        """Lists all properties of attribute form."""
        return super().to_dict(camel_case=camel_case)

    def to_dict(self, camel_case=True) -> dict:
        return {
            key: val
            for key, val in self.list_properties(camel_case).items()
            if key
            not in [
                'id',
                'abbreviation',
                'type',
                'subtype',
                'ext_type',
                'date_created',
                'date_modified',
                'version',
                'owner',
                'icon_path',
                'view_media',
                'ancestors',
                'certified_info',
                'acg',
                'acl',
                'target_info',
                'formType',
            ]
        }

    def local_alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        display_format: Optional[DisplayFormat] = None,
        data_type: Optional[DataType] = None,
        expressions: Optional[list[FactExpression]] = None,
        alias: Optional[str] = None,
        lookup_table: Optional[SchemaObjectReference] = None,
        child_forms: Optional[list[FormReference]] = None,
        geographical_role: Optional[GeographicalRole] = None,
        time_role: Optional[TimeRole] = None,
        is_form_group: Optional[bool] = None,
        is_multilingual: Optional[bool] = None,
    ):
        """Make changes to a local copy of AttributeForm. In order to change
        an AttributeForm object on the server, use `Attribute.alter_form()`
        method of a corresponding Attribute object.
        """
        properties = filter_params_for_func(self.local_alter, locals())

        for key, val in delete_none_values(properties, recursion=True).items():
            self.__setattr__(name=key, value=val)

    def get_fact_expression(self, id: str) -> FactExpression:
        """Get a fact expression of a local instance of
        AttributeForm object.

        Args:
            id: id of the fact expression
        """
        expressions = [expr for expr in self.expressions if expr.id == id]
        if not expressions:
            raise ValueError(
                f"AttributeForm with id {self.id} does not contain "
                f"Fact Expressions with id {id}."
            )
        return expressions[0]

    # Attribute form expressions management
    def _add_fact_expression(self, expression: FactExpression):
        """
        Internal method that affects ONLY LOCAL AttributeForm object.
        To make changes on the server, use `Attribute._add_fact_expression()`
        method of a corresponding Attribute object.
        Add expression to the form.
        Args:
            form_id: id of the form to which the expression is to be added,
            expression: the expression that is to be added,
        """
        self.verify_is_simple_form(method_name_if_error='Adding expression')
        self.expressions.append(expression)

    def _remove_fact_expression(
        self,
        fact_expression_id: str,
        new_lookup_table: Optional[SchemaObjectReference] = None,
    ):
        """
        Internal method that affects ONLY LOCAL AttributeForm object.
        To make changes on the server, use `Attribute._remove_fact_expression()`
        method of a corresponding Attribute object.
        Remove expression from the form. If the expressions left are
        not using lookup table assigned to the form, provide new lookup
        table for the form.

        Args:
            form_id: id of the form from which the expression is to be removed,
            fact_expression_id: id of the expression that is to be removed,
            new_lookup_table: new lookup table of the form. Used if the
                expression that is going to be removed used the current one
        """
        self.verify_is_simple_form(method_name_if_error='Removing expression')
        if len(self.expressions) == 1:
            exception_handler(
                "You can't remove the only expression "
                f"from the attribute form with id: {self.id}.",
                ValueError,
            )
        ex_postition = next(
            (i for i, ex in enumerate(self.expressions) if ex.id == fact_expression_id),
            None,
        )
        if ex_postition is None:
            exception_handler(
                f"The form with id: {self.id}, don't use "
                f"expression with id: {fact_expression_id}.",
                ValueError,
            )
        removed_ex = self.expressions.pop(ex_postition)

        # AttributeForm can't use `lookup_table` that is not used in any of
        # its expressions.
        if self.lookup_table in removed_ex.tables:
            if len(self.expressions) == 1 and len(self.expressions[0].tables) == 1:
                self.lookup_table = self.expressions[0].tables[0]
            elif new_lookup_table is None:
                exception_handler(
                    "Please provide new lookup_table for the attribute form "
                    f"with id: {self.id}.",
                    AttributeError,
                )
            else:
                self.lookup_table = new_lookup_table

    def _alter_expression(
        self,
        fact_expression_id: str,
        expression: Optional[Expression] = None,
        tables: Optional[list[SchemaObjectReference]] = None,
    ):
        """Internal method that affects ONLY LOCAL AttributeForm object.
        To make changes on the server, use `Attribute.alter_fact_expression()`
        method of a corresponding Attribute object.

        Args:
            fact_expression_id: id of the fact expression that is to be altered,
            expression: new expressions of the fact expression
            tables: new tables of the fact expression
        """
        self.verify_is_simple_form(method_name_if_error='Altering expressions')
        try:
            fact_expr = next(
                expr for expr in self.expressions if expr.id == fact_expression_id
            )
            fact_expr.local_alter(expression, tables)
        except StopIteration:
            exception_handler(
                f"Attribute Form: {self.id} does not have Fact Expression"
                f" with ID: {fact_expression_id}.",
                ValueError,
            )

    def verify_is_simple_form(self, method_name_if_error: str):
        """Verify whether the form is a simple form and not a group form"""
        if self.is_form_group:
            exception_handler(
                f"{method_name_if_error} is not available for the form group.",
                AttributeError,
            )

    def is_referenced_by(self, form_reference: FormReference) -> bool:
        """Check if attribute form is referenced in `form_reference`."""
        if (self.id is not None and self.id == form_reference.id) or (
            self.name is not None and self.name == form_reference.name
        ):
            return True
        return False



from functools import wraps
import inspect
import logging
from typing import Any, Callable, Union

from requests.adapters import Response

from mstrio import config
from mstrio.api.exceptions import MstrException, PartialSuccess, Success
from mstrio.utils.helper import get_default_args_from_func, response_handler


def get_args_and_bind_values(func: Callable[[Any], Any], *args, **kwargs):
    signature = inspect.signature(func)
    return signature.bind(*args, **kwargs).arguments


logger = logging.getLogger(__name__)


class ErrorHandler:
    """An easy-to-use class decorator designed to replace
    logic responsible for displaying the error message in API wrappers.

    Attributes:
        err_msg(str): error message to be displayed in case of error

    Usage:
        to replace the code below

          if not response.ok:
            if error_msg is None:
              error_msg = f'Error deleting Datasource Login with ID {id}'
            response_handler(response, error_msg)
          return response

        use the decorator in a following way

           @ErrorHandler(err_msg='Error deleting Datasource Login with ID {id}')
           def func(connection, id):
              ...

        the strings in curly braces will be replaced
        with the appropriate values if they appear in the function arguments
    """

    def __init__(self, err_msg: str):
        self._err_msg = err_msg

    def __call__(self, func: Callable[[Any], Any]):
        @wraps(func)
        def inner(*args, **kwargs):
            response = func(*args, **kwargs)
            error_msg = (
                kwargs.get("error_msg") if kwargs.get("error_msg") else self._err_msg
            )
            if not response.ok:
                handler_kwargs = self._get_resp_handler_kwargs(kwargs)
                error_msg = self._replace_with_values(error_msg, func, *args, **kwargs)
                response_handler(response, error_msg, **handler_kwargs)
            return response

        return inner

    @staticmethod
    def _replace_with_values(err_msg: str, func: Callable[[Any], Any], *args, **kwargs):
        all_args = get_args_and_bind_values(func, *args, **kwargs)
        for arg in all_args:
            arg_name, arg_value = arg, all_args[arg]
            err_msg = err_msg.replace(f'{{{arg_name}}}', str(arg_value))
        return err_msg

    @staticmethod
    def _get_resp_handler_kwargs(decorated_func_kwargs):
        default_args = get_default_args_from_func(response_handler)
        for arg in default_args:
            default_args[arg] = decorated_func_kwargs.get(arg, default_args[arg])
        return default_args


def bulk_operation_response_handler(
    response: Response, unpack_value: str = None
) -> Union[PartialSuccess, Success, MstrException]:
    """Handle partial success and other statuses from bulk operation."""
    response_body = response.json()
    if response.ok and unpack_value:
        response_body = response_body[unpack_value]

    if response.status_code == 200:
        err = Success(response_body)
    elif response.status_code == 207:
        err = PartialSuccess(response_body)
    else:
        err = MstrException(response_body)

    if config.verbose:
        logger.error(err)
    return err



# flake8: noqa
# isort: off
from .package import (
    Package,
    PackageConfig,
    PackageContentInfo,
    PackageImport,
    PackageSettings,
)
from .migration import (
    bulk_full_migration,
    bulk_migrate_package,
    Migration,
    MigrationStatus,
)

# isort: on



import json
from typing import Optional

from requests.adapters import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error getting schedules list.')
def list_schedules(connection, fields=None, error_msg=None):
    """Get list of a schedules.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    response = connection.get(
        url=f'{connection.base_url}/api/schedules', params={'fields': fields}
    )
    if response.ok:
        # Fix for incorrect 'eventId' (expecting 'id')
        event_based_in_list = False
        response_json = response.json()
        for schedule in response_json['schedules']:
            if 'event' in schedule.keys():
                schedule['event']['id'] = schedule['event'].pop('eventId')
                event_based_in_list = True
        if event_based_in_list:
            response.encoding, response._content = 'utf-8', json.dumps(
                response_json
            ).encode('utf-8')

    return response


@ErrorHandler(err_msg='Error getting schedule {id} information.')
def get_schedule(connection, id, fields=None, error_msg=None):
    """Get information of a specific schedule by `schedule_id`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        schedule_id(str): ID of the schedule
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    response = connection.get(
        url=f'{connection.base_url}/api/schedules/{id}', params={'fields': fields}
    )
    if response.ok:
        # Fix for incorrect 'eventId' (expecting 'id')
        response_json = response.json()
        if 'event' in response_json.keys():
            response_json['event']['id'] = response_json['event'].pop('eventId')
            response.encoding, response._content = 'utf-8', json.dumps(
                response_json
            ).encode('utf-8')

    return response


@ErrorHandler(err_msg='Error getting schedule information.')
def create_schedule(connection, body, fields=None, error_msg=None):
    """Create a new schedule using data from `body`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        body(dict): Dictionary containing data used for creating schedule.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    # id to eventId conversion - API Problem
    if 'event' in body.keys():
        body['event']['eventId'] = body['event'].pop('id')

    response = connection.post(
        url=f'{connection.base_url}/api/schedules', json=body, params={'fields': fields}
    )
    if response.ok:
        # Fix for incorrect 'eventId' (expecting 'id')
        response_json = response.json()
        if 'event' in response_json.keys():
            response_json['event']['id'] = response_json['event'].pop('eventId')
            response.encoding, response._content = 'utf-8', json.dumps(
                response_json
            ).encode('utf-8')

    return response


@ErrorHandler(err_msg='Error getting schedule {id} information.')
def update_schedule(connection, id, body, fields=None, error_msg=None):
    """Alter a schedule specified by `schedule_id`, using data from `body`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        schedule_id(str): ID of the schedule
        body(dict): Dictionary containing data used for replacing schedule.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    # id to eventId conversion - API Problem
    if 'event' in body.keys():
        body['event']['eventId'] = body['event'].pop('id')

    response = connection.put(
        url=f'{connection.base_url}/api/schedules/{id}',
        json=body,
        params={'fields': fields},
    )
    if response.ok:
        # Fix for incorrect 'eventId' (expecting 'id')
        response_json = response.json()
        if 'event' in response_json.keys():
            response_json['event']['id'] = response_json['event'].pop('eventId')
        response.encoding, response._content = 'utf-8', json.dumps(
            response_json
        ).encode('utf-8')

    return response


@ErrorHandler(err_msg='Error getting schedule {id} information.')
def delete_schedule(connection, id, fields=None, error_msg=None):
    """Delete a schedule specified by `schedule_id`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the schedule
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.delete(
        url=f'{connection.base_url}/api/schedules/{id}', params={'fields': fields}
    )


@ErrorHandler(err_msg='Error getting schedule information')
def get_contents_schedule(
    connection: "Connection", project_id: str, body: dict, fields: Optional[str] = None
) -> Response:
    """Get a list of available schedules for a list of contents within a given
        project.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id(str): Field to filter on project id (exact match),
        body: List of contents
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.post(
        url=f'{connection.base_url}/api/schedules/results',
        headers={'X-MSTR-ProjectID': project_id},
        params={'fields': fields},
        json=body,
    )



"""This is the demo script to show how administrator can manage subscriptions.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.distribution_services import (
    CacheType,
    CacheUpdateSubscription,
    Content,
    EmailSubscription,
    FileSubscription,
    FTPSubscription,
    HistoryListSubscription,
    list_subscriptions,
    Subscription,
    SubscriptionManager
)
from mstrio.api.users import create_address_v2, get_addresses
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

# Create connection based on workstation data
CONN = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
EMAIL_SUBSCRIPTION_NAME = $email_subscription_name
# see distribution_services/subscription/content.py for available options
FORMAT_TYPE = $format_type
EMAIL_SUBJECT = $email_subject
SCHEDULE_ID = $schedule_id
CONTENT_ID = $content_id
# see distribution_services/subscription/content.py for available options
CONTENT_TYPE = $content_type
RECIPIENT_ID = $recipient_id
RECIPIENT_ID_2 = $recipient_id_2
RECIPIENT_ID_3 = $recipient_id_3

# Create an email subscription
EMAIL_SUB = EmailSubscription.create(
    connection=CONN,
    name=EMAIL_SUBSCRIPTION_NAME,
    project_name=PROJECT_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=CONTENT_TYPE,
        personalization=Content.Properties(format_type=FORMAT_TYPE)
    ),
    schedules=[SCHEDULE_ID],
    recipients=[RECIPIENT_ID],
    email_subject=EMAIL_SUBJECT
)

# Define variables which can be later used in a script
FILE_SUBSCRIPTION_NAME = $file_subscription_name
FILE_NAME = $file_name
ZIP_FILE_NAME = $zip_file_name
ZIP_PASSWORD = $zip_password

# Create a file subscription
FILE_SUB = FileSubscription.create(
    connection=CONN,
    name=FILE_SUBSCRIPTION_NAME,
    project_name=PROJECT_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=CONTENT_TYPE,
        personalization=Content.Properties(format_type=FORMAT_TYPE)
    ),
    schedules=[SCHEDULE_ID],
    recipients=[RECIPIENT_ID_3],
    filename=FILE_NAME,
    zip_password_protect=True,
    zip_filename=ZIP_FILE_NAME,
    zip_password=ZIP_PASSWORD
)

# Define variables which can be later used in a script
FTP_SUBSCRIPTION_NAME = $ftp_subscription_name
FTP_ADDRESS_NAME = $ftp_address_name
PHYSICAL_ADDRESS = $physical_address
DEVICE_ID = $ftp_device_id
# FTP sub needs a user with FTP Device address
# Body for FTP Device address
BODY = {
        'name': FTP_ADDRESS_NAME,
        'physicalAddress': PHYSICAL_ADDRESS,
        'deliveryType': 'FTP',
        'deviceId': DEVICE_ID,
        'isDefault': False
    }
# Create FTP address for user
create_address_v2(connection=CONN, id=RECIPIENT_ID_3, body=BODY)
# Create an ftp subscription
FTP_SUB = FTPSubscription.create(
    connection=CONN,
    name=FTP_SUBSCRIPTION_NAME,
    project_name=PROJECT_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=CONTENT_TYPE,
        personalization=Content.Properties(format_type=FORMAT_TYPE)
    ),
    schedules=[SCHEDULE_ID],
    recipients=[RECIPIENT_ID_3],
    filename=FILE_NAME,
    zip_password_protect=True,
    zip_filename=ZIP_FILE_NAME,
    zip_password=ZIP_PASSWORD
)

# Define variables which can be later used in a script
HISTORY_LIST_SUBSCRIPTION_NAME = $history_list_subscription_name
DELIVERY_EXPIRATION_DATE = $delivery_expiration_date

# Create a history list subscription
HL_SUB = HistoryListSubscription.create(
    connection=CONN,
    name=HISTORY_LIST_SUBSCRIPTION_NAME,
    project_name=PROJECT_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=CONTENT_TYPE,
        personalization=Content.Properties(format_type=FORMAT_TYPE)
    ),
    schedules=[SCHEDULE_ID],
    recipients=[RECIPIENT_ID_3],
    do_not_create_update_caches=True,
    overwrite_older_version=True,
    re_run_hl=True
)

# Define variables which can be later used in a script
CACHE_SUBSCRIPTION_NAME = $cache_subscription_name
DELIVERY_EXPIRATION_TIME_ZONE = $delivery_expiration_time_zone

# Create a cache update subscription
CACHE_UPDATE_SUB = CacheUpdateSubscription.create(
    connection=CONN,
    project_name=PROJECT_NAME,
    name=CACHE_SUBSCRIPTION_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=CONTENT_TYPE,
        personalization=Content.Properties(format_type=Content.Properties.FormatType.PDF),
    ),
    schedules=[SCHEDULE_ID],
    delivery_expiration_date=DELIVERY_EXPIRATION_DATE,
    delivery_expiration_timezone=DELIVERY_EXPIRATION_TIME_ZONE,
    send_now=True,
    recipients=[RECIPIENT_ID_3],
    cache_cache_type=CacheType.RESERVED
)  # see distribution_services/subscription/delivery.py for available options

# Define variables which can be later used in a script
CACHE_SUBSCRIPTION_NEW_NAME = $cache_subscription_new_name
OWNER_ID = $owner_id

# Change name and owner of a subscription
CACHE_UPDATE_SUB.alter(name=CACHE_SUBSCRIPTION_NEW_NAME, owner_id=OWNER_ID)

# List a particular subscription type, for example Cache Update
CACHE_UPDATE_SUBS = [
    sub for sub in list_subscriptions(CONN, project_name=PROJECT_NAME)
    if isinstance(sub, CacheUpdateSubscription)
]

# Initialize manager for subscriptions on a chosen project
sub_mngr = SubscriptionManager(connection=CONN, project_name=PROJECT_NAME)
# Get all subscriptions from the given project (it is possible in two ways)
all_subs = list_subscriptions(connection=CONN, project_name=PROJECT_NAME)
all_subs = sub_mngr.list_subscriptions()

# Execute/delete subscriptions by passing theirs ids or Subscription objects
sub_mngr.execute([EMAIL_SUB.id])
sub_mngr.delete([EMAIL_SUB, FTP_SUB, FILE_SUB, CACHE_UPDATE_SUB], force=True)

# List available recipients of the subscription for the given content (default
# delivery type is an email)
sub_mngr.available_recipients(content_id=CONTENT_ID, content_type=CONTENT_TYPE)

# Get a single subscription
sub = Subscription(connection=CONN, subscription_id=HL_SUB.id, project_name=PROJECT_NAME)
# List all recipients of the given subscription and all available for this
# subscription
sub.recipients
sub.available_recipients()

# Define variables which can be later used in a script
RECIPIENT_ID_4 = $recipient_id_4

# Add/remove recipient(s) with given id(s)
sub.add_recipient(recipients=[RECIPIENT_ID_4])
sub.remove_recipient(recipients=[RECIPIENT_ID_4])

# As an alternate to subscription manager you can execute or delete a subscription
# directly from the subscription itself
sub.execute()
sub.delete(force=True)

# Define variables which can be later used in a script
USER_TO_REMOVE_ID = $user_to_remove_id
ADMIN_USER_ID = $admin_user_id

# Replace a user with an admin in all of its subscriptions (e.g. when user exits
# company)
for s in sub_mngr.list_subscriptions(to_dictionary=False):
    if USER_TO_REMOVE_ID in [r['id'] for r in s.recipients]:
        s.add_recipient(recipients=ADMIN_USER_ID)
        s.remove_recipient(recipients=USER_TO_REMOVE_ID)



from json.decoder import JSONDecodeError
import logging
import os

from mstrio.utils.helper import exception_handler

logger = logging.getLogger(__name__)


def print_url(
    response, *args, **kwargs  # NOSONAR required for hook to session objects
):
    """Response hook to print url for debugging."""
    logger.debug(response.url)


def save_response(
    response, *args, **kwargs  # NOSONAR required for hook to session objects
):
    """Response hook to save REST API responses to files structured by the API
    family."""
    import json
    from pathlib import Path

    if response.status_code != 204:
        # Generate file name
        base_path = Path(__file__).parents[2] / 'tests/resources/auto-api-responses/'
        url = response.url.rsplit('api/', 1)[1]
        temp_path = url.split('/')
        file_name = '-'.join(temp_path[1:]) if len(temp_path) > 1 else temp_path[0]
        file_name = f'{file_name}-{response.request.method}'
        file_path = base_path if len(temp_path) == 1 else base_path / temp_path[0]
        path = str(file_path / file_name)

        # Create target directory & all intermediate directories if don't exists
        if not os.path.exists(str(file_path)):
            os.makedirs(file_path)
            print("Directory ", file_path, " created ")
        else:
            print("Directory ", file_path, " already exists")

        # Dump the response to JSON and Pickle
        # with open(path + '.pkl', 'wb') as f:
        #     pickle.dump(response, f)
        with open(path + '.json', 'w') as f:
            try:
                json.dump(response.json(), f)
            except JSONDecodeError:
                exception_handler(
                    "Could not decode response. Skipping creating JSON file.", Warning
                )



from typing import Optional

import requests

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error while creating the package holder')
def create_package_holder(
    connection: Connection,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Create a new in-memory metadata package holder.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.post(
        url=f'{connection.base_url}/api/packages',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error while updating the package holder with id: {id}')
def update_package_holder(
    connection: Connection,
    body: dict,
    id: str,
    project_id: Optional[str] = None,
    prefer: str = "respond-async",
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Fill the content of the in-memory metadata package holder per supplied
    specification. Currently, it's only supported when the holder is empty.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        body (dict): dictionarized PackageConfig object (with `to_dict()`)
        id (str): ID of the package to be updated
        prefer (str, optional): API currently just supports asynchronous mode,
        not support synchronous mode, so header parameter ‘Prefer’ must be set
        to ‘respond-async’ in your request. Defaults to "respond-async".
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.put(
        url=f'{connection.base_url}/api/packages/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        json=body,
    )


@ErrorHandler(err_msg='Error while downloading the package with id: {id}')
def download_package(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Download a package binary.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): ID of the package to be downloaded.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        url=f'{connection.base_url}/api/packages/{id}/binary',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error while uploading the package with id: {id}')
def upload_package(
    connection: Connection,
    id: str,
    file: bytes,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Upload package to sandbox directly.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): ID of the package to be uploaded.
        file (bytes): package in a format of a binary string.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.put(
        url=f'{connection.base_url}/api/packages/{id}/binary',
        headers={'X-MSTR-ProjectID': project_id},
        files={'file': file},
    )


@ErrorHandler(err_msg='Error while getting the package holder with id: {id}')
def get_package_holder(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    show_content: bool = True,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Get definition of a package, including package status and its detail
    content.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): ID of the package to be retrieved.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        show_content (bool, optional): Show package content or not. Defaults to
        False.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id

    return connection.get(
        url=f'{connection.base_url}/api/packages/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'showContent': show_content},
    )


@ErrorHandler(err_msg='Error while deleting the package holder with id: {id}')
def delete_package_holder(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    prefer: str = 'respond-async',
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Delete the in-memory metadata package holder, releasing associated
    Intelligence Server resources.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): ID of the package to be deleted.
        prefer (str, optional): API currently just supports asynchronous mode,
        not support synchronous mode, so header parameter ‘Prefer’ must be set
        to ‘respond-async’ in your request. Defaults to "respond-async".
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """

    project_id = project_id if project_id is not None else connection.project_id
    return connection.delete(
        url=f'{connection.base_url}/api/packages/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
    )


@ErrorHandler(
    err_msg='Error while creating the import for package holder with id: {id}'
)
def create_import(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    generate_undo: bool = False,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Create a package import process.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): ID of the package for which import process will be
        created.
        generate_undo (bool, optional): Generate undo package or not. Defaults
        to False.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """

    # TODO: Change to a parameter when any other values are supported
    prefer = 'respond-async'

    project_id = project_id if project_id is not None else connection.project_id
    return connection.post(
        url=f'{connection.base_url}/api/packages/imports',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={'packageId': id, 'generateUndo': generate_undo},
    )


@ErrorHandler(err_msg='Error while getting the import with id: {id}')
def get_import(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Get result of a package import process.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): Import process ID.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        url=f'{connection.base_url}/api/packages/imports/{id}',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error while deleting the import with id: {id}')
def delete_import(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Closes an existing import process previously created.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): Import process ID.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    # TODO: Change to a parameter when any other values are supported
    prefer = 'respond-async'

    project_id = project_id if project_id is not None else connection.project_id
    return connection.delete(
        url=f'{connection.base_url}/api/packages/imports/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
    )


@ErrorHandler(err_msg='Error while creating the undo for import with id: {id}')
def create_undo(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> requests.Response:
    """Download undo package binary for this import process.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server.
        id (str): Import process ID.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        url=f'{connection.base_url}/api/packages/imports/{id}/undoPackage/binary',
        headers={'X-MSTR-ProjectID': project_id},
    )



from mstrio.connection import Connection
from mstrio.api import drivers as drivers_api
from mstrio.utils.helper import rename_dict_keys


REST_ATTRIBUTES_MAP = {'enabled': 'isEnabled'}


def get_all(connection: Connection) -> list[dict]:
    """Get information for all drivers.

    Args:
        connection: MicroStrategy REST API connection object

    Returns:
        list of dict representing driver objects
    """
    drivers = drivers_api.get_drivers(connection).json()['drivers'].values()
    return [rename_dict_keys(item, REST_ATTRIBUTES_MAP) for item in drivers]


def get(connection: Connection, id: str) -> dict:
    """Get driver by a specific ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the driver

    Returns:
        dict representing driver object
    """
    data = drivers_api.get_driver(connection, id).json()
    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)


def update(
    connection: Connection,
    id: str,
    body: dict,
) -> dict:
    """Update a driver.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the driver
        body: Driver update info.

    Returns:
        dict representing driver object
    """
    data = drivers_api.update_driver(connection, id, body).json()
    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)



# flake8: noqa
from .content_cache import ContentCache
from .datasets import *
from .document import Document, list_documents, list_documents_across_projects
from .dossier import (
    ChapterPage,
    Dossier,
    list_dossiers,
    list_dossiers_across_projects,
    PageSelector,
    PageVisualization,
    VisualizationSelector,
)
from .library import Library
from .report import list_reports, Report



from typing import Optional, TYPE_CHECKING

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession


@ErrorHandler(err_msg='Error while creating the folder.')
def create_folder(
    connection: Connection,
    name: str,
    parent_id: str,
    description: Optional[str] = None,
    project_id: Optional[str] = None,
):
    """Create a folder.

    Note:
        When `project_id` is provided then folder will be created in this
        project. Otherwise it will be created in a project selected within
        `connection` object.

    Args:
        connection: MicroStrategy REST API connection object
        name (string): name of folder to create
        parent_id (string): id of folder in which new folder will be created
        description (string, optional): description of folder to create
        project_id (string, optional): id of project

    Returns:
        Complete HTTP response object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    body = {
        "name": name,
        "description": description,
        "parent": parent_id,
    }
    return connection.post(
        url=connection.base_url + '/api/folders',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg='Error while deleting folder with ID: {id}.')
def delete_folder(connection: Connection, id: str, project_id: Optional[str] = None):
    """Delete complete folder.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project

    Returns:
        Complete Future object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.delete(
        url=f"{connection.base_url}/api/folders/{id}",
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error while listing folders.')
def list_folders(
    connection: Connection,
    project_id: Optional[str] = None,
    offset: int = 0,
    limit: int = 5000,
    error_msg: Optional[str] = None,
):
    """Get a list of folders.

    Args:
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f"{connection.base_url}/api/folders",
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit},
    )


def list_folders_async(
    future_session: "FuturesSession",
    connection: Connection,
    project_id: Optional[str] = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Get a list of folders asynchronously.

    Args:
        future_session(object): `FuturesSession` object to call MicroStrategy
            REST Server asynchronously
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    url = f"{connection.base_url}/api/folders"
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error while getting contents of a folder with ID: {id}.')
def get_folder_contents(
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    offset: int = 0,
    limit: int = 5000,
    error_msg: Optional[str] = None,
):
    """Get contents of a folder.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        url=f"{connection.base_url}/api/folders/{id}",
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit},
    )


def get_folder_contents_async(
    future_session: "FuturesSession",
    connection: Connection,
    id: str,
    project_id: Optional[str] = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Get contents of a folder asynchronously.

    Args:
        future_session(object): `FuturesSession` object to call MicroStrategy
            REST Server asynchronously
        connection: MicroStrategy REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    url = f"{connection.base_url}/api/folders/{id}"
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error while getting contents of a pre-defined folder.')
def get_predefined_folder_contents(
    connection: Connection,
    folder_type: int,
    project_id: Optional[str] = None,
    offset: int = 0,
    limit: int = 5000,
    error_msg: Optional[str] = None,
):
    """Get contents of a pre-defined folder.

    Args:
        connection: MicroStrategy REST API connection object
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f"{connection.base_url}/api/folders/preDefined/{folder_type}",
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit},
    )


def get_predefined_folder_contents_async(
    future_session: "FuturesSession",
    connection: Connection,
    folder_type: int,
    project_id: Optional[str] = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Get contents of a pre-defined folder.

    Args:
        future_session(object): `FuturesSession` object to call MicroStrategy
            REST Server asynchronously
        connection: MicroStrategy REST API connection object
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    url = f"{connection.base_url}/api/folders/preDefined/{folder_type}"
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error while getting contents of My Personal Objects folder.')
def get_my_personal_objects_contents(
    connection: Connection, project_id: Optional[str] = None
):
    """Get contents of My Personal Objects folder.

    Args:
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project

    Returns:
         Complete HTTP response object.
    """
    return connection.get(
        url=f"{connection.base_url}/api/folders/myPersonalObjects",
        headers={'X-MSTR-ProjectID': project_id},
    )



# flake8: noqa
from .attribute import *
from .fact import *
from .helpers import *
from .schema_management import (
    SchemaLockStatus,
    SchemaLockType,
    SchemaManagement,
    SchemaTask,
    SchemaTaskStatus,
    SchemaUpdateType,
)
from .table import *
from .transformation import *
from .user_hierarchy import *



import logging
from typing import List, Optional, Union

from packaging import version

from mstrio import config
from mstrio.api import events, objects
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_events(
    connection: Connection, to_dictionary: bool = False, limit: int = None, **filters
) -> Union[List["Event"], List[dict]]:
    """List event objects or event dictionaries. Optionally filter list.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        to_dictionary(bool, optional): if True, return event as
            list of dicts
        limit(int, optional): maximum number of Events returned.
        **filters: Available filter parameters:
            ['name', 'id', 'description', 'acg']
    """
    _objects = helper.fetch_objects(
        connection=connection,
        api=events.list_events,
        limit=limit,
        filters=filters,
        dict_unpack_value='events',
    )

    if to_dictionary:
        return _objects
    else:
        return [Event.from_dict(source=obj, connection=connection) for obj in _objects]


@class_version_handler('11.3.0100')
class Event(Entity, DeleteMixin):
    """Class representation of MicroStrategy Event object.

    Attributes:
        connection: A MicroStrategy connection object
        name: Event name
        id: Event ID
        description: Event descriptions
    """

    _PATCH_PATH_TYPES = {'name': str, 'description': str}
    _OBJECT_TYPE = ObjectTypes.SCHEDULE_EVENT
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
        ): objects.get_object_info,
    }
    _API_DELETE = staticmethod(events.delete_event)
    _API_PATCH = {('name', 'description'): (events.update_event, 'put')}

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the Event object, populates it with I-Server data.
        Specify either `id` or `name`. When `id` is provided (not `None`),
        `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Event ID
            name: Event name
        """
        self._API_GETTERS[('id', 'name', 'description')] = (
            events.get_event
            if version.parse(connection.web_version) >= version.parse('11.3.0200')
            else objects.get_object_info
        )

        if id is None and name is None:
            raise AttributeError(
                "Please specify either 'name' or 'id' parameter in the constructor."
            )
        if id is None:
            objects_info = list_events(connection, name=name, to_dictionary=True)
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no event with the given name: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id, name=name)

    def trigger(self):
        """Trigger the Event"""
        response = events.trigger_event(self.connection, self.id)
        if response.ok and config.verbose:
            logger.info(
                f"Event '{self.name}' with ID : '{self.id}' has been triggered."
            )
        return response.ok

    @classmethod
    def create(
        cls, connection: Connection, name: str, description: Optional[str] = None
    ) -> "Event":
        """Create an Event

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            name: Name of the new Event
            description: Description of the new Event
        """
        body = helper.delete_none_values(
            {
                "name": name,
                "description": description,
            },
            recursion=True,
        )
        response = events.create_event(connection, body)
        return cls.from_dict(response.json(), connection)

    def alter(self, name: Optional[str] = None, description: Optional[str] = None):
        """Alter the Event's properties

        Args:
            name: New name for the Event
            description: New description for the Event
        """
        args = helper.delete_none_values(
            {
                "name": name,
                "description": description,
            },
            recursion=True,
        )
        self._alter_properties(**args)
        if config.verbose:
            logger.info(f"Updated subscription '{self.name}' with ID: {self.id}.")



"""List all the columns for all available tables"""
from typing import Optional, Union

from mstrio.api import tables
from mstrio.connection import Connection, get_connection
from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.api_helpers import async_get
from mstrio.utils.wip import module_wip, WipLevels

module_wip(globals(), level=WipLevels.PREVIEW)


def list_table_columns(
    connection: Connection, table: Optional[Union[SchemaObjectReference, str]] = None
) -> dict:
    """List all the columns for a given table, if tables is not specified,
    list all the columns for all available tables.

    Args:
        connection: MicroStrategy connection object returned
            by `connection.Connection()`
        table: SchemaObjectReference of a table or table id

    Returns:
        dictionary of all the columns for a given table, or for all tables
        if table is not specified e.g.
        {'TABLE_NAME':[
            SchemaObjectReference(sub_type='column', object_id='1111',
                                  name='column_name'),
            SchemaObjectReference(sub_type='column', object_id='2222',
                                  name='other_column_name')
        ]}

    """

    def unpack(table_res):
        columns_list_dict = table_res['physicalTable']['columns']
        return [
            SchemaObjectReference(
                col['information']['subType'],
                col['information']['objectId'],
                col['information']['name'],
            )
            for col in columns_list_dict
        ]

    if table:
        table_id = table if isinstance(table, str) else table.object_id
        table_res = tables.get_table(
            connection, table_id, project_id=connection.project_id
        ).json()
        return {table_res['information']['name']: unpack(table_res)}
    else:
        table_res = tables.get_tables(
            connection, project_id=connection.project_id
        ).json()
        ids = [tab['information']['objectId'] for tab in table_res['tables']]
        tables_async = async_get(tables.get_table_async, connection, ids)
        return {table['information']['name']: unpack(table) for table in tables_async}


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# list all the columns for all available tables
list_table_columns(conn)



from typing import List, TYPE_CHECKING, Union

from mstrio.api import datasources
from mstrio.utils import helper
from mstrio.utils.entity import EntityBase, ObjectTypes
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@method_version_handler('11.3.0000')
def list_available_dbms(
    connection: "Connection", to_dictionary: bool = False, limit: int = None, **filters
) -> Union[List["Dbms"], List[dict]]:
    """List all available database management systems (DBMSs) objects or dicts.
    Optionally filter the DBMS by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'version']

    Examples:
        >>> list_available_dbms(connection)
    """
    return Dbms._list_available_dbms(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


@class_version_handler('11.3.0000')
class Dbms(EntityBase):
    """Object representation of MicroStrategy Database management system (DBMS)

    Attributes:
        name: Database management system (DBMS) name.
        id: Database management system (DBMS) ID.
        type: Database management system (DBMS) type.
        version: Database management system (DBMS) version.
    """

    _OBJECT_TYPE = ObjectTypes.NONE
    _FROM_DICT_MAP = {}  # map attributes to Enums and Composites
    _DBMS_CACHE = set()

    def __init__(
        self,
        connection: "Connection",
        name: str = None,
        id: str = None,
        type: str = None,
        version: str = None,
    ):
        """Initialize Dbms object."""
        if name is None and id is None:
            helper.exception_handler(
                "Please specify either 'name' or 'id' parameter in the constructor.",
                exception_type=ValueError,
            )
        temp_dbms = None

        for dbms in self._DBMS_CACHE:
            if name == dbms.name or id == dbms.id:
                temp_dbms = dbms
                super().__init__(
                    connection=connection,
                    object_id=temp_dbms.id,
                    name=temp_dbms.name,
                    type=temp_dbms.type,
                    version=temp_dbms.version,
                )

        if not temp_dbms:
            if id:
                temp_dbms = Dbms._list_available_dbms(
                    connection, id=id, to_dictionary=True
                )
            elif name:
                temp_dbms = Dbms._list_available_dbms(
                    connection, name=name, to_dictionary=True
                )

            if temp_dbms:
                [temp_dbms] = temp_dbms
                super().__init__(
                    connection=connection,
                    object_id=temp_dbms["id"],
                    name=temp_dbms["name"],
                    type=temp_dbms["type"],
                    version=temp_dbms["version"],
                )
            else:
                identifier = name if name else id
                raise ValueError(f"There is no Dbms: '{identifier}'")

    def _init_variables(self, **kwargs) -> None:
        """Set object attributes by providing keyword args."""
        super()._init_variables(**kwargs)
        self._version = kwargs.get("version")

    @classmethod
    def _list_available_dbms(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: int = None,
        **filters,
    ) -> Union[List["Dbms"], List[dict]]:
        objects = helper.fetch_objects(
            connection=connection,
            api=datasources.get_available_dbms,
            limit=None,
            filters=None,
        )
        cls._DBMS_CACHE.update(
            [cls.from_dict(source=obj, connection=connection) for obj in objects]
        )

        if limit:
            objects = objects[:limit]
        if filters:
            objects = helper.filter_list_of_dicts(objects, **filters)
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @property
    def version(self) -> str:
        return self._version



from typing import Optional, TYPE_CHECKING

from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.sessions import FuturesSessionWithRenewal

if TYPE_CHECKING:
    from mstrio.connection import Connection


@unpack_information
@ErrorHandler("Error getting the table with ID: {id}")
def get_table(
    connection: "Connection",
    id: str,
    project_id: Optional[str] = None,
    changeset_id: Optional[str] = None,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get a detailed definition for a specified table.

    Args:
        connection (object): MicroStrategy REST API connection object
        id (str): Table ID
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected(),
        project_id = connection.project_id

    return connection.get(
        url=f"{connection.base_url}/api/model/tables/{id}",
        headers={"X-MSTR-ProjectID": project_id, "X-MSTR-Changeset": changeset_id},
        params={"fields": fields},
    )


@unpack_information
@ErrorHandler("Error listing tables")
def get_tables(
    connection: "Connection",
    project_id: Optional[str] = None,
    changeset_id: Optional[str] = None,
    limit: int = None,
    offset: int = 0,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get a list of all tables.

    Args:
        connection (object): MicroStrategy REST API connection object
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use None for no limit.
            Default is None.
        offset (int, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default is 0.
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected(),
        project_id = connection.project_id

    return connection.get(
        url=f"{connection.base_url}/api/model/tables",
        headers={"X-MSTR-ProjectID": project_id, "X-MSTR-Changeset": changeset_id},
        params={"limit": limit, "offset": offset, "fields": fields},
    )


@unpack_information
@ErrorHandler("Error updating the table with ID: {id}")
def patch_table(
    connection: "Connection",
    id: str,
    body: dict,
    column_merge_option: Optional[str] = None,
    fields: Optional[dict] = None,
    error_msg: Optional[str] = None,
):
    """Update a detailed definition for a specified table in the changeset

    Args:
        connection (object): MicroStrategy REST API connection object
        id (str): Table ID
        body (dict): Table update body
        column_merge_option (str, optional): Defines a column merge option
            Available values: 'reuse_any', 'reuse_compatible_data_type',
            'reuse_matched_data_type'.
            If 'reuse_any', updates the column body type to use the most
                recent column definition.
            If 'reuse_compatible_data_type', updates the column body type
                to use the body type with the largest precision or scale.
            If 'reuse_matched_data_type', renames the column in newly added
                table to allow it to have different body types.
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.patch(
            url=f"{connection.base_url}/api/model/tables/{id}",
            headers={"X-MSTR-MS-Changeset": changeset_id},
            json=body,
            params={"columnMergeOption": column_merge_option, "fields": fields},
        )


@unpack_information
@ErrorHandler("Error creating the table")
def post_table(
    connection: "Connection",
    data: dict,
    check_secondary_data_source_table: Optional[bool] = None,
    column_merge_option: Optional[str] = None,
    table_prefix_option: Optional[str] = None,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Update a detailed definition for a specified table in the changeset

    Args:
        connection (object): MicroStrategy REST API connection object
        data (dict): Table creation data
        check_secondary_data_source_table (bool, optional):
            Available values: 'true', 'false'
            If 'true', finds compatible tables in the project.
                If a compatible table is found, the compatible table object
                information is returned. If no table is found,
                a new table is created.
            If 'false', a new table is created.
        column_merge_option (str, optional): Defines a column merge option
            Available values: 'reuse_any', 'reuse_compatible_data_type',
            'reuse_matched_data_type'.
            If 'reuse_any', updates the column data type to use the most
                recent column definition.
            If 'reuse_compatible_data_type', updates the column data type
                to use the data type with the largest precision or scale.
            If 'reuse_matched_data_type', renames the column in newly added
                table to allow it to have different data types.
        table_prefix_option (str, optional): Define the table prefix
            Available values: 'none', 'add_default_prefix', 'add_namespace'.
            If 'none', do not set table prefix.
            If 'add_default_prefix', applies the default prefix setting on
                warehouse catalog.
            If 'add_namespace', create a prefix same with namespace.
        fields(list, optional): Comma separated top-level field whitelist.
            This allows client to selectively retrieve part of the
            response model.
        error_msg (str, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            url=f"{connection.base_url}/api/model/tables",
            headers={"X-MSTR-MS-Changeset": changeset_id},
            json=data,
            params={
                "checkSecondaryDataSourceTable": "true"
                if check_secondary_data_source_table
                else "false",
                "columnMergeOption": column_merge_option,
                "tablePrefixOption": table_prefix_option,
                "fields": fields,
            },
        )


@ErrorHandler("Error fetching available Warehouse Tables")
def get_available_warehouse_tables(
    connection: "Connection",
    datasource_id: str,
    namespace_id: str,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    if project_id is None:
        connection._validate_project_selected(),
        project_id = connection.project_id

    url = (
        f"{connection.base_url}/api/datasources/{datasource_id}/catalog/"
        f"namespaces/{namespace_id}/tables"
    )
    return connection.get(
        url,
        headers={
            "X-MSTR-ProjectID": project_id,
        },
    )


def get_table_async(
    session: FuturesSessionWithRenewal,
    connection: "Connection",
    id: str,
    changeset_id: Optional[str] = None,
    project_id: str = None,
    fields=None,
):
    """Get definition of a single table by id, using FuturesSessionWithRenewal

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a table
        changeset_id: ID of a changeset
        project_id: Id of a project
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.
    Return:
        Complete Future object.
    """
    return session.get(
        url=f"{connection.base_url}/api/model/tables/{id}",
        headers={"X-MSTR-MS-Changeset": changeset_id, "X-MSTR-ProjectID": project_id},
        params={"fields": fields},
    )


def get_available_warehouse_tables_async(
    session: FuturesSessionWithRenewal,
    connection: "Connection",
    datasource_id: str,
    namespace_id: str,
):
    return session.get(
        f"{connection.base_url}/api/datasources/{datasource_id}/catalog/namespaces/"
        f"{namespace_id}"
        "/tables",
        headers={
            "X-MSTR-ProjectID": connection.project_id,
        },
    )



from typing import Optional, TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession

    from mstrio.connection import Connection


@ErrorHandler(err_msg='Error listing Contact Groups.')
def get_contact_groups(
    connection: "Connection",
    offset: int = 0,
    limit: int = 1000,
    error_msg: Optional[str] = None,
):
    """Get a list of all contact groups that the authenticated user has access
        to.

    Args:
        connection: MicroStrategy REST API connection object
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 1000.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/contactGroups"
    params = {'offset': offset, 'limit': limit}
    return connection.get(url=url, params=params)


def get_contact_groups_async(
    future_session: "FuturesSession",
    connection: "Connection",
    offset: int = 0,
    limit: int = 1000,
):
    """Get a list of all contact groups that the authenticated user has access
        to.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        connection: MicroStrategy REST API connection object
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 1000.

    Returns:
        Complete Future object.
    """
    url = f'{connection.base_url}/api/contactGroups'
    params = {'offset': offset, 'limit': limit}
    future = future_session.get(url, params=params)
    return future


@ErrorHandler(err_msg='Error creating Contact Group.')
def create_contact_group(
    connection: "Connection", body: dict, error_msg: Optional[str] = None
):
    """Create a new contact group.

    Args:
        connection: MicroStrategy REST API connection object
        body: Contact group creation body
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/contactGroups"
    return connection.post(url=url, json=body)


@ErrorHandler(err_msg='Error getting Contact Group with ID {id}')
def get_contact_group(
    connection: "Connection", id: str, error_msg: Optional[str] = None
):
    """Get contact group by a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact group
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/contactGroups/{id}"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error updating Contact Group with ID {id}')
def update_contact_group(
    connection: "Connection", id: str, body: dict, error_msg: Optional[str] = None
):
    """Update a contact group.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact group
        body: Contact group update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/contactGroups/{id}"
    return connection.put(url=url, json=body)


@ErrorHandler(err_msg='Error deleting Contact Group with ID {id}')
def delete_contact_group(
    connection: "Connection", id: str, error_msg: Optional[str] = None
):
    """Delete a contact group.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the contact group
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/contactGroups/{id}"
    return connection.delete(url=url)



from dataclasses import dataclass
from enum import auto
from typing import List, Optional, Type, TYPE_CHECKING

from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import camel_to_snake, Dictable

from .dynamic_date_time import DynamicDateTimeStructure
from .expression import Expression

if TYPE_CHECKING:
    from mstrio.connection import Connection


class ParameterType(AutoName):
    """Enumeration constant indicating type of parameter"""

    CONSTANT = auto()
    OBJECT_REFERENCE = auto()
    EXPRESSION = auto()
    PROMPT = auto()
    ARRAY = auto()
    DYNAMIC_DATE_TIME = auto()


class VariantType(AutoName):
    """Enumeration constant indicating type of value of a variant object"""

    INT32 = auto()
    INT64 = auto()
    DATE = auto()
    TIME = auto()
    BOOLEAN = auto()
    STRING = auto()
    DOUBLE = auto()


@dataclass
class PredicateParameter(Dictable):
    """Base class for parameters"""

    _TYPE = None

    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict(camel_case)
        result[camel_case and 'parameterType' or 'parameter_type'] = self._TYPE.value

        return result

    @staticmethod
    def dispatch(
        source, connection: Optional['Connection'] = None
    ) -> Type['PredicateParameter']:
        """Returns an appropriate PredicateParameter object from the
            provided source

        Args:
            source: object that specifies PredicateParameter that will be
                returned
            connection (optional): MicroStrategy connection object returned
            by `connection.Connection()`"""
        data = camel_to_snake(source)
        parameter_type = ParameterType(data.pop('parameter_type'))
        cls = PREDICATE_PARAMETER_MAP[parameter_type]

        return cls.from_dict(data, connection, to_snake_case=False)


@dataclass
class AttributeElement(Dictable):
    """This object represents a concrete element.
    A concrete element is a metadata representation of an element
    that exists in some data source.

    Since an external data source can be modified without using
    the MicroStrategy Platform, the existence of a concrete element in
    the metadata does not mean that the element described currently exists.
    It means that it existed at some point.

    Attributes:
        display: Human readable string to display to describe the element to
            a reader
        element_id: Opaque string value that identifies the element to
            the service.

            The client should make no assumption about the format of this
            string, and should not attempt to parse it (even though forms of
            the element are likely to be visibly concatenated into it). This
            string will remain valid even if the current client is using a
            different locale to the locale of the client that created the
            element.

            When creating an element, the client should copy this string from
            other fields of the element browse result. If a "terse" element id
            is used, the service will need to know the attribute in order to
            decode the element id.
        attribute: Optionally the element may specify the attribute that it uses

            Note that in most cases elements are used in a context in which the
            attribute value is implied (for example the element is part of an
            IN-list of elements for some attribute). if the attribute is
            implied there is no need to redundantly repeat it on each element
            in the collection.

    """

    _FROM_DICT_MAP = {
        'attribute': SchemaObjectReference,
    }
    display: str
    element_id: str
    attribute: Optional[SchemaObjectReference] = None


@dataclass
class Variant(Dictable):
    """The typed value used in the predicates

    Attributes:
        type: Identifies the type of `value` attribute
        value: The value, encoded as a string in a standard,
            locale-neutral, format.
    """

    _FROM_DICT_MAP = {
        'type': VariantType,
    }

    type: VariantType
    value: str


@dataclass
class FunctionProperty(Dictable):
    """Additional properties associated with a function.

    Attributes:
        name: function's property name
        value: function's property value
    """

    name: str
    value: Variant


@dataclass
class ConstantParameter(PredicateParameter):
    """Object used to describe a parameter that appears in a filter predicate
    and whose value is constant.

    Attributes:
        constant: representation of constant as Variant object
    """

    _TYPE = ParameterType.CONSTANT
    _FROM_DICT_MAP = {
        'constant': Variant,
    }
    constant: Variant


@dataclass
class ObjectReferenceParameter(PredicateParameter):
    """Object that describe a parameter that appears in a filter predicate and
    whose value is an object reference.

    For example, a parameter of this type would be used by a filter predicate
    that compares the value of two metrics. This would be modelled by using a
    metric predicate (containing one of the two metrics); this parameter would
    then be used to record the other metric.

    Attributes:
        target: reference to object
    """

    _TYPE = ParameterType.OBJECT_REFERENCE
    _FROM_DICT_MAP = {
        'target': SchemaObjectReference,
    }
    target: SchemaObjectReference


@dataclass
class ExpressionParameter(PredicateParameter):
    """Object used to describe a parameter that appears in a filter predicate
    and whose value is given as a custom expression.

    This predicate parameter option is used for when the end user wants to
    provide a custom expression for the predicate. We would expect the parser
    to be used to modify the expression.

    Attributes:
        expression: custom expression, instance of Expression class
    """

    _TYPE = ParameterType.EXPRESSION
    _FROM_DICT_MAP = {
        'expression': Expression,
    }
    expression: Expression


@dataclass
class PromptParameter(PredicateParameter):
    """Object used to describe a parameter that appears in a filter predicate
    and whose value will be obtained on execution by resolving a prompt object.

    The prompt could be any type of prompt that returns a simple value: integer
    prompt, double prompt, big-decimal prompt, string prompt or date prompt.
    The prompt should be consistent with the data type of the value to be
    compared with the prompt.

    Attributes:
        prompt: reference to prompt object
    """

    _TYPE = ParameterType.PROMPT
    _FROM_DICT_MAP = {
        'prompt': SchemaObjectReference,
    }
    prompt: SchemaObjectReference


@dataclass
class ConstantArrayParameter(PredicateParameter):
    """Object used to describe a parameter that appears in a filter predicate
    and whose value is a list of constants.

    This kind of parameter is very specific. It is used for the IN and NOT_IN
    comparison functions. These functions compare one value with an array of
    constant values. Note that all the values in the array must have the
    same type.

    Attributes:
        constants_type: Identifies the type of values in `constants` attribute
        constants: list of constant values, each of which is encoded as a string
            in a standard, locale-neutral representation.
            Their common datatype is given by constants_type.

    """

    _TYPE = ParameterType.ARRAY
    _FROM_DICT_MAP = {
        'constants_type': VariantType,
    }
    constants_type: VariantType
    constants: List[str]


@dataclass
class DynamicDateTimeParameter(PredicateParameter):
    """Object used to describe a parameter that appears in a filter predicate
    and whose value is a dynamic date/time.

    Attributes:
        dynamic_date_time: Object that represents a date/time.
    """

    _TYPE = ParameterType.DYNAMIC_DATE_TIME
    _FROM_DICT_MAP = {
        'dynamic_date_time': DynamicDateTimeStructure,
    }
    dynamic_date_time: DynamicDateTimeStructure


PREDICATE_PARAMETER_MAP = {
    ParameterType.ARRAY: ConstantArrayParameter,
    ParameterType.CONSTANT: ConstantParameter,
    ParameterType.DYNAMIC_DATE_TIME: DynamicDateTimeParameter,
    ParameterType.EXPRESSION: ExpressionParameter,
    ParameterType.OBJECT_REFERENCE: ObjectReferenceParameter,
    ParameterType.PROMPT: PromptParameter,
}



from enum import Enum
from itertools import chain
from typing import Type


class AutoName(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self.lower()

    @classmethod
    def has_value(cls, value):
        return value in set(item.value for item in cls)


class AutoUpperName(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self


class AutoCapitalizedName(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self.capitalize()


def __get_enum_helper(
    obj, enum: Type[Enum] = Enum, get_value: bool = False
) -> str | Type[Enum] | None:
    """Helper function for `get_enum` and `get_enum_val`."""
    if obj is None:
        return obj

    if isinstance(obj, Enum) and isinstance(obj, enum):
        obj = obj.value if get_value else obj
    elif isinstance(obj, (str, int)):
        validate_enum_value(obj, enum)
        obj = obj if get_value else enum(obj)
    else:
        raise TypeError(f"Incorrect type. Value should be of type: {enum}.")

    return obj


def get_enum(obj, enum: Type[Enum] = Enum) -> Type[Enum] | None:
    """Safely get enum from enum or str."""
    return __get_enum_helper(obj, enum)


def get_enum_val(obj, enum: Type[Enum] = Enum) -> str | None:
    """Safely extract value from enum or str."""
    return __get_enum_helper(obj, enum, True)


def validate_enum_value(obj, enum):
    """Validate provided value. If not correct,
    error message with possible options will be displayed.
    """
    from mstrio.utils.helper import exception_handler

    possible_values = (
        [[e.value for e in item] for item in enum]
        if isinstance(enum, tuple)
        else [e.value for e in enum]
    )
    err_msg = f"Incorrect enum value '{obj}'. Possible values are {possible_values}"
    if obj not in list(chain(possible_values)):
        exception_handler(err_msg, exception_type=ValueError)
